@@ -1,6 +1,6 @@
 package SVN::Mirror::Ra;
 @ISA = ('SVN::Mirror');
-$VERSION = '0.65';
+$VERSION = '0.67';
 use strict;
 use SVN::Core;
 use SVN::Repos;
@@ -428,14 +428,18 @@ The structure of mod_lists:
             $local_path = "$self->{target_path}/$svn_lpath";
         }
 
+	my $local_rev = -1;
+	unless ($item->copyfrom_rev == -1) {
+	    $local_rev = $self->find_local_rev
+		($item->copyfrom_rev, $self->{rsource_uuid});
+	}
 	# XXX: the logic of the code here is a mess!
         my ($action, $rpath, $rrev, $lrev) =
             @$href{qw/action remote_path remote_rev local_rev local_path/} =
                 ( $item->action,
                   $item->copyfrom_path,
                   $item->copyfrom_rev,
-                  ($item->copyfrom_rev == -1) ?
-                    -1 : $self->find_local_rev ($item->copyfrom_rev, $self->{rsource_uuid}) || undef,
+		  $local_rev,
                   $local_path,
                 );
 	# workaround fsfs remoet_path inconsistencies
@@ -465,13 +469,12 @@ The structure of mod_lists:
         @$href{qw/local_source_path source_node_kind/} =
             ( $src_lpath, $source_node_kind );
 
-	# XXX: do we still need the reanchor case?
-	# Also, the loop should not reached here if changed path is
+	# XXX: the loop should not reached here if changed path is
 	# not interesting to us, skip them at the beginning the the loop
         if ( $_ eq $self->{rsource_path} or
 	     index ("$_/", "$self->{rsource_path}/") == 0 ) {
             $editor->{mod_lists}{$svn_lpath} = $href;
-        } elsif (defined $lrev && $lrev != -1 && $href->{action} eq 'A' &&
+        } elsif ($rrev != -1 && $href->{action} eq 'A' &&
 		 index ($self->{rsource_path}, "$_/") == 0) {
 	    # special case for the parent of the anchor is copied.
 	    my $reanchor = $self->{rsource_path};
@@ -484,15 +487,31 @@ The structure of mod_lists:
     }
 
     unless (keys %{$editor->{mod_lists}}) {
-        print "Skipped revision $rev.\n";
+	my $root = $editor->open_root($self->{headrev});
+	$editor->change_dir_prop ($root, svm => undef);
+	$editor->close_directory($root);
+	$editor->close_edit;
     } else {
-        my @mod_list = keys %{$editor->{mod_lists}};
+        my @mod_list = sort keys %{$editor->{mod_lists}};
+	# mark item as directory that we are sure about.
+	# do not use !isdir for deciding the item is _not_ a directory.
+	for my $parent (@mod_list) {
+	    for (@mod_list) {
+		next if $parent eq $_;
+		if (index ("$_/", "$parent/") == 0) {
+		    $editor->{mod_lists}{$parent}{isdir} = 1;
+		    last;
+		}
+	    }
+	}
         if ( ($self->{skip_to} && $self->{skip_to} <= $rev) ||
 	     grep { my $href = $editor->{mod_lists}{$_};
                     !( ( ($href->{action} eq 'A' || $href->{action} eq 'R')
-                         && defined $href->{local_rev}
-                         && $href->{local_rev} != -1
-                         && $href->{source_node_kind} == $SVN::Node::dir)
+                         && ((defined $href->{local_rev}
+			      && $href->{local_rev} != -1
+			      && $href->{source_node_kind} == $SVN::Node::dir)
+			     || ($href->{isdir})
+			    ))
                        || $href->{action} eq 'D' )
                 } @mod_list ) {
 	    my $pool = SVN::Pool->new_default_sub;
@@ -513,7 +532,7 @@ The structure of mod_lists:
 
             $edit->open_root ($self->{headrev});
 
-            foreach (sort @mod_list) {
+            foreach (@mod_list) {
                 my $href = $editor->{mod_lists}{$_};
                 my $action = $href->{action};
 
@@ -522,8 +541,13 @@ The structure of mod_lists:
                 }
 
                 if ($action eq 'A' || $action eq 'R') {
-                    $edit->copy_directory ($_, $href->{local_source_path},
-                                           $href->{local_rev});
+		    if (defined $href->{local_rev} && $href->{local_rev} != -1) {
+			$edit->copy_directory ($_, $href->{local_source_path},
+					       $href->{local_rev});
+		    }
+		    else {
+			$edit->add_directory($_);
+		    }
                     $edit->close_directory ($_);
 		}
 	    }
@@ -636,12 +660,19 @@ sub run {
 
     print "Retrieving log information from $startrev to $endrev\n";
 
+    my $firsttime = 1;
     eval {
     $ra->get_log ([''], $startrev, $endrev,
 		  ($SVN::Core::VERSION ge '1.2.0') ? (0) : (),
 		  1, 1,
 		  sub {
 		      my ($paths, $rev, $author, $date, $msg, $pool) = @_;
+		      # for the first time, skip_to might not hit
+		      # active revision in the tree. adjust to make it so.
+		      if ($firsttime) {
+			  $self->{skip_to} = $rev if defined $self->{skip_to};
+			  $firsttime = 0;
+		      }
 		      # move the anchor detection stuff to &mirror ?
 		      if (defined $self->{skip_to} && $rev <= $self->{skip_to}) {
 			  # XXX: get the logs for skipped changes
@@ -738,6 +769,10 @@ sub open_directory {
     print "MirrorEditor::open_directory($path)\n" if $debug;
     return undef unless $pb;
 
+    if ( ($self->_in_modified_list($path) || '') eq 'R') {
+	return $self->add_directory($path, $pb, undef, -1, $pool);
+    }
+
     my $dir_baton = $self->SUPER::open_directory ($path, $pb,
                                                   $self->{mirror}{headrev},
                                                   $pool);
@@ -756,6 +791,12 @@ sub open_file {
     my ($self,$path,$pb,undef,$pool) = @_;
     print "MirrorEditor::open_file($path)\n" if $debug;
     return undef unless $pb;
+    my $action = $self->_in_modified_list ($path);
+    if ($self->_under_latest_copypath ($path)
+	&& $self->_is_under_true_copy ($path)
+	&& !$action) {
+	return undef;
+    }
     $self->{opening} = $path;
     return $self->SUPER::open_file ($path, $pb,
 				    $self->{mirror}{headrev}, $pool);
@@ -1174,7 +1215,10 @@ sub delete_entry {
     my ($self, $path, $rev, $pb, $pool) = @_;
     print "MirrorEditor::delete_entry($path, $rev)\n" if $debug;
     return unless $pb;
-
+    if ($self->_under_latest_copypath($path)) {
+	my $action = $self->_in_modified_list($path) || '';
+	return unless $action eq 'D' || $action eq 'R';
+    }
     $self->SUPER::delete_entry ($path, $self->{mirror}{headrev},
 				$pb, $pool);
 }
